using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;

using Microsoft.Xna.Framework;

namespace Syntriax.Engine.Physics2D.Primitives;

public record Line(Vector2 From, Vector2 To)
{
    public Vector2 Direction => Vector2.Normalize(To - From);
    public float Length => (From - To).Length();
    public float LengthSquared => (From - To).LengthSquared();

    public LineEquation LineEquation
    {
        get
        {
            Vector2 slopeVector = To - From;
            float slope = slopeVector.Y / slopeVector.X;

            float yOffset = From.Y - (slope * From.X);

            return new LineEquation(slope, yOffset);
        }
    }

    public bool Intersects(Vector2 point)
        => Resolve(point.X).ApproximatelyEquals(point);

    public float GetT(Vector2 point)
    {
        float fromX = MathF.Abs(From.X);
        float toX = MathF.Abs(To.X);
        float pointX = MathF.Abs(point.X);

        float min = MathF.Min(fromX, toX);
        float max = MathF.Max(fromX, toX) - min;

        pointX -= min;

        return pointX / max;
    }

    public bool Exist(List<Vector2> vertices)
    {
        for (int i = 0; i < vertices.Count - 1; i++)
        {
            Vector2 vertexCurrent = vertices[i];
            Vector2 vertexNext = vertices[i];
            if (From == vertexCurrent && To == vertexNext) return true;
            if (From == vertexNext && To == vertexCurrent) return true;
        }

        Vector2 vertexFirst = vertices[0];
        Vector2 vertexLast = vertices[^1];
        if (From == vertexFirst && To == vertexLast) return true;
        if (From == vertexLast && To == vertexFirst) return true;
        return false;
    }

    public float IntersectionParameterT(Line other)
        => ((other.From.X - From.X) * (To.Y - From.Y) - (other.From.Y - From.Y) * (To.X - From.X)) /
            ((other.To.Y - other.From.Y) * (To.X - From.X) - (other.To.X - other.From.X) * (To.Y - From.Y));


    public Vector2 Resolve(float x)
        => new Vector2(x, LineEquation.Resolve(x));

    public Vector2 ClosestPointTo(Vector2 point)
    {
        // Convert edge points to vectors
        var edgeVector = new Vector2(To.X - From.X, To.Y - From.Y);
        var pointVector = new Vector2(point.X - From.X, point.Y - From.Y);

        // Calculate the projection of pointVector onto edgeVector
        float t = (pointVector.X * edgeVector.X + pointVector.Y * edgeVector.Y) / (edgeVector.X * edgeVector.X + edgeVector.Y * edgeVector.Y);

        // Clamp t to the range [0, 1] to ensure the closest point is on the edge
        t = Math.Max(0, Math.Min(1, t));

        // Calculate the closest point on the edge
        float closestX = From.X + t * edgeVector.X;
        float closestY = From.Y + t * edgeVector.Y;

        return new Vector2((float)closestX, (float)closestY);
    }

    public Vector2 IntersectionPoint(Line other)
        => Vector2.Lerp(From, To, IntersectionParameterT(other));

    public bool Intersects(Line other)
    {
        int o1 = PhysicsMath.Orientation(From, To, other.From);
        int o2 = PhysicsMath.Orientation(From, To, other.To);
        int o3 = PhysicsMath.Orientation(other.From, other.To, From);
        int o4 = PhysicsMath.Orientation(other.From, other.To, To);

        if (o1 != o2 && o3 != o4)
            return true;

        if (o1 == 0 && PhysicsMath.OnSegment(From, other.From, To)) return true;
        if (o2 == 0 && PhysicsMath.OnSegment(From, other.To, To)) return true;
        if (o3 == 0 && PhysicsMath.OnSegment(other.From, From, other.To)) return true;
        if (o4 == 0 && PhysicsMath.OnSegment(other.From, To, other.To)) return true;

        return false;
    }

    public bool Intersects(Line other, [NotNullWhen(returnValue: true)] out Vector2? point)
    {
        point = null;

        bool result = Intersects(other);

        if (result)
            point = IntersectionPoint(other);

        return result;
    }
}
